

import os
import boto3
import typeguard
from botocore.exceptions import ClientError
from os import listdir
from os.path import isfile, join

from benchmark_runner.common.clouds.shared.s3.s3_operations_exceptions import S3FileNotUploaded, S3FileNotDownloaded, S3FileNotDeleted, S3KeyNotCreated, S3FileNotExist, S3FailedCreatePresingedURL
from benchmark_runner.main.environment_variables import environment_variables


class S3Operations:
    """ This class is responsible for S3 operations """

    def __init__(self, region_name: str = '', endpoint_url: str = None, aws_access_key_id: str = None, aws_secret_access_key: str = None):
        # environment variables
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        # must add region for pytest
        if region_name:
            self.__region = region_name
            self.__endpoint_url = endpoint_url
            self.__aws_access_key_id = aws_access_key_id
            self.__aws_secret_access_key = aws_secret_access_key
        else:
            self.__region = self.__environment_variables_dict.get('region_name', '')
            # must be None for pytest
            self.__endpoint_url = self.__environment_variables_dict.get('endpoint_url', None)
            self.__aws_access_key_id = self.__environment_variables_dict.get('access_key_id', '')
            self.__aws_secret_access_key = self.__environment_variables_dict.get('secret_access_key', '')
        self.__s3_client = boto3.client(service_name='s3',
                                        region_name=self.__region,
                                        endpoint_url=self.__endpoint_url,
                                        aws_access_key_id=self.__aws_access_key_id,
                                        aws_secret_access_key=self.__aws_secret_access_key)

    @typeguard.typechecked
    def upload_file(self, file_name_path: str, bucket: str, key: str, upload_file: str):
        """
        This method uploads file/object to s3
        :param file_name_path:'/home/user/test.txt'
        :param bucket:'benchmark'
        :param key:'test-data'
        :param upload_file:'test.txt'
        :return:
        """
        try:
            self.__s3_client.upload_file(Filename=file_name_path,
                                         Bucket=bucket,
                                         Key=f'{key}/{upload_file}',
                                         ExtraArgs={'ServerSideEncryption': 'AES256'})
        except ClientError:
            raise
        except Exception:
            raise S3FileNotUploaded

    @typeguard.typechecked
    def download_file(self, bucket: str, key: str, download_file: str, file_name_path: str):
        """
        This method downloads file/object from s3
        :param bucket:'benchmark'
        :param key:'logs/ec2-idle/2021/01/19/18'
        :param download_file: 'test.txt'
        :param file_name_path:'D:\\Performance\\Projects\\py-image-service\\data\\rt_results\\test.txt'
        :return:
        """
        try:
            if download_file:
                self.__s3_client.download_file(Bucket=bucket, Key=f'{key}/{download_file}', Filename=file_name_path)
            else:
                self.__s3_client.download_file(Bucket=bucket, Key=key, Filename=file_name_path)
        except ClientError:
            raise
        except Exception:
            raise S3FileNotDownloaded

    @typeguard.typechecked
    def delete_file(self, bucket: str, key: str, file_name: str):
        """
        This method deletes file/object from s3
        :param bucket:'benchmark'
        :param key:'test-data'
        :param file_name: 'test.txt'
        :return:
        """
        try:
            self.__s3_client.delete_object(Bucket=bucket, Key=f'{key}/{file_name}')
        except ClientError:
            raise
        except Exception:
            raise S3FileNotDeleted

    @typeguard.typechecked
    def delete_folder(self, bucket: str, key: str):
        """
        This method deletes folder/key from s3
        :param bucket:'benchmark'
        :param key:'framework/test'
        :return:
        """
        try:
            objects_to_delete = self.__s3_client.list_objects(Bucket=bucket, Prefix=key)
            delete_keys = {
                'Objects': [{'Key': k} for k in [obj['Key'] for obj in objects_to_delete.get('Contents', [])]]}
            if delete_keys['Objects']:
                self.__s3_client.delete_objects(Bucket=bucket, Delete=delete_keys)
        except ClientError:
            raise
        except Exception:
            raise S3FileNotDeleted

    @typeguard.typechecked
    def create_folder(self, bucket: str, key: str):
        """
        This method creates folder/key is s3
        :param bucket:'benchmark'
        :param key:'framework/test'
        :return:
        """
        try:
            self.__s3_client.put_object(Bucket=bucket, Key=key)
        except ClientError:
            raise
        except Exception:
            raise S3KeyNotCreated

    @typeguard.typechecked
    def file_exist(self, bucket: str, key: str, file_name: str):
        """
        This method checks if file/object exists in s3
        :param bucket:'benchmark'
        :param key:'framework/test'
        :param file_name:'file.txt'
        :return:
        """
        try:
            response = self.__s3_client.list_objects_v2(Bucket=bucket, Prefix=key)
            if response.get('Contents'):
                for item in response['Contents']:
                    if file_name in item['Key']:
                        return True
            return False
        except ClientError:
            raise
        except Exception:
            raise S3FileNotExist

    @typeguard.typechecked
    def folder_exist(self, bucket: str, key: str):
        """
        This method checks if folder/key exists in s3
        :param bucket:'benchmark'
        :param key:'framework/test'
        :return:
        """
        try:
            response = self.__s3_client.list_objects_v2(Bucket=bucket, Prefix=key)
            if response.get('Contents'):
                return True
            return False
        except ClientError:
            raise
        except Exception:
            raise S3KeyNotCreated

    @typeguard.typechecked
    def upload_objects(self, local_source: str, s3_target: str):
        """
        This method uploads local data folder to s3 target path
        :param local_source: local data folder i.e. '/home/user/'
        :param s3_target: target s3 path i.e. 'data_store/calc_image_data/'
        :return:
        """
        try:
            if '/' in s3_target:
                targets = s3_target.split('/')
                bucket = targets[0]
                key = '/'.join(targets[1:])
            else:
                bucket = s3_target
                key = ''

            files = [f for f in listdir(local_source) if isfile(join(local_source, f))]
            for file in files:
                filename = os.path.join(local_source, file)
                self.upload_file(file_name_path=filename, bucket=bucket, key=key, upload_file=file)

        except ClientError as err:
            raise
        except Exception:
            raise S3FileNotUploaded

    @typeguard.typechecked
    def download_objects(self, s3_target: str, local_source: str):
        """
        This method downloads from s3 target to local data folder
        :param local_source: local data folder i.e. '/home/user/
        :param s3_target: target s3 path i.e. 'data_store/calc_image_data/'
        :return:
        """
        files = []
        try:
            if '/' in s3_target:
                targets = s3_target.split('/')
                bucket = targets[0]
                key = '/'.join(targets[1:])
            else:
                bucket = s3_target
                key = ''

            response = self.__s3_client.list_objects_v2(Bucket=bucket, Prefix=key)
            if response.get('Contents'):
                for item in response['Contents']:
                    if item['Key'].split('/')[-1]:
                        files.append(item['Key'].split('/')[-1])
                    else:
                        files.append(item['Key'])

            for file in files:
                file_name = os.path.join(local_source, file)
                self.download_file(bucket=bucket, key=key, download_file=file, file_name_path=file_name)
        except ClientError:
            raise
        except Exception:
            raise S3FileNotDownloaded

    @typeguard.typechecked
    def generate_presigned_url(self, bucket: str, key: str, file_name: str):
        """
        This method generate presigned url for specific uploaded object, default 7 days
        :param bucket:'benchmark'
        :param key:'logs/test-data'
        :param file_name:'file.txt'
        :return:
        """
        try:
            return self.__s3_client.generate_presigned_url('get_object',
                                                           Params={'Bucket': bucket, 'Key': f'{key}/{file_name}'},
                                                           ExpiresIn=604800)
        except ClientError:
            raise
        except Exception:
            raise S3FailedCreatePresingedURL




import copy
from benchmark_runner.main.environment_variables import environment_variables


class TemporaryEnvironmentVariables:
    """
    Allow temporary replacement of environment variables for testing purposes.
    Intended to be used as

    with TemporaryEnvironmentVariables():
        environment_variables.environment_variables_dict[variable] = temporary_value
        ...

    """

    def __init__(self):
        __saved_environment__ = None

    def __enter__(self):
        temporary_environment = copy.deepcopy(environment_variables.environment_variables_dict)
        self.__saved_environment__ = copy.deepcopy(environment_variables.environment_variables_dict)
        environment_variables.environment_variables_dict = temporary_environment

    def __exit__(self, exception_type, exception_value, traceback):
        if self.__saved_environment__:
            environment_variables.environment_variables_dict = self.__saved_environment__




# display df
from IPython.display import clear_output

# display bokeh
from bokeh.plotting import figure, show
from bokeh.io import output_notebook, output_file
from bokeh.models import HoverTool, Label, ColumnDataSource, LabelSet, Circle, NumeralTickFormatter


class VisualizeWorkloadOperations:
    """
    This class for visualize workloads operations
    """

    WIDTH = 1200
    HEIGHT = 600

    def __init__(self):
        self.clear_output()

    @staticmethod
    def clear_output():
        """
        This method clean output
        @return:
        """
        # clean latest bokeh output
        clear_output(wait=True)

    def compare_run_results(self, result1: dict, result2: dict, legend_label1: str, legend_label2: str, **workload_data: dict):
        """
        This method compare between 2 results
        @param result1: thread results in dictionary
        @param result2: thread results in dictionary
        @param legend_label1: first bar legend
        @param legend_label2: second bar legend
        @return:
        """
        title = workload_data.get('title')
        x_axis_label = workload_data.get('x_axis_label')
        y_axis_label = workload_data.get('y_axis_label')

        # Convert values to integers
        values1 = [float(value) for value in result1.values()]
        values2 = [float(value) for value in result2.values()]

        metrics = list(result1.keys())

        width = 0.3
        x = [i for i in range(len(metrics))]
        offset = 0.1  # Adjust the offset for x-axis shift
        x_middle = [i + width / 2 + offset for i in x]  # Adjust x-coordinates for centering and shift

        p = figure(title=title, x_range=metrics, width=self.WIDTH, height=self.HEIGHT)
        p.xaxis.axis_label = x_axis_label  # Add x-axis title
        p.yaxis.axis_label = y_axis_label  # Add y-axis title

        r1 = p.vbar(x=x_middle, top=values1, width=width,
                    line_color='white', fill_color='green', alpha=0.8,
                    legend_label=legend_label1, name=legend_label2)

        r2 = p.vbar(x=[i + width for i in x_middle], top=values2, width=width,
                    line_color='white', fill_color='blue', alpha=0.8,
                    legend_label=legend_label2, name=legend_label1)

        p.xgrid.grid_line_color = None
        p.legend.location = 'top_right'
        p.legend.title = 'Versions'

        p.x_range.range_padding = 0.1  # Adjust the padding between bars

        tooltips_1 = [('Version', legend_label1), (f'{y_axis_label}', '@top{0,0.00}')]
        tooltips_2 = [('Version', legend_label2), (f'{y_axis_label}', '@top{0,0.00}')]

        hover_1 = HoverTool(renderers=[r1], tooltips=tooltips_1)
        hover_2 = HoverTool(renderers=[r2], tooltips=tooltips_2)

        p.add_tools(hover_1, hover_2)  # Add the HoverTools to the figure

        # Add y-values on top of each bar
        for i, value in enumerate(values1):
            label = Label(x=x_middle[i], y=value, text=f'{value:,.2f}', text_font_size='8pt', text_color='black',
                          text_baseline='bottom', text_align='center')
            p.add_layout(label)

        for i, value in enumerate(values2):
            label = Label(x=x_middle[i] + width, y=value, text=f'{value:,.2f}', text_font_size='8pt',
                          text_color='black',
                          text_baseline='bottom', text_align='center')
            p.add_layout(label)

        p.yaxis.formatter = NumeralTickFormatter(format='0,')  # Format y-axis tick labels with commas

        # Percentage line

        # Calculate the percentage difference between result1 and result2
        # latency: Lower is better
        if 'latency'.upper() in [key.upper().upper() for key in result1.keys()]:
            percentage_diff = [(v1 - v2) / v2 * 100 if v2 != 0 else 0 for v1, v2 in zip(values1, values2)]
        else:  # Other: Higher is better
            percentage_diff = [(v2 - v1) / v1 * 100 if v1 != 0 else 0 for v1, v2 in zip(values1, values2)]
        percentage_diff = [round(diff, 2) for diff in percentage_diff]  # Round to 2 decimal places

        # Calculate the x-coordinates for the dots
        x_middle = [i + width / 2 + offset for i in x]

        # Calculate the maximum value from result1 and result2 dictionaries
        min_value = min(min(values1), min(values2))

        # Calculate the y-coordinate for the dots
        y_middle = [max(values1[i], values2[i]) + min_value for i in range(len(metrics))]

        # Add labels and dots for percentage difference
        for i, diff in enumerate(percentage_diff):
            x_pos = x_middle[i] + width / 4  # Adjust the x-coordinate to shift the label to the right
            y_pos = y_middle[i] + 0.1  # Adjust the y-coordinate to move the labels above the line
            color = 'red' if diff < 0 else 'green'  # Set label and dot color based on the sign of the difference
            label = Label(x=x_pos, y=y_pos, text=f'{diff}%', text_font_size='14pt', text_color=color,
                          text_baseline='bottom', text_align='center')
            p.add_layout(label)

            # Adjust the y-coordinate to place the dots below the labels
            dot_y = y_pos - 0.1
            dot_x = x_pos + width / 4  # Adjust the x-coordinate to shift the dots to the right

            # Change dot color based on the sign of the difference
            dot_color = 'red' if diff < 0 else 'green'
            dot_source = ColumnDataSource(data=dict(x=[dot_x], y=[dot_y], color=[dot_color]))
            p.circle(x='x', y='y', size=8, fill_color='color', line_color='color', source=dot_source)

        # Display the graph
        output_notebook()
        show(p)




import os
import time
from multiprocessing import Process

from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.common.elasticsearch.elasticsearch_exceptions import ElasticSearchDataNotUploaded
from benchmark_runner.workloads.workloads_operations import WorkloadsOperations


class VdbenchVM(WorkloadsOperations):
    """
    This class runs vdbench vm
    """
    START_STAMP = '@@~@@START-WORKLOAD@@~@@'
    END_STAMP = '@@~@@END-WORKLOAD@@~@@'

    def __init__(self):
        super().__init__()
        self.__name = ''
        self.__workload_name = ''
        self.__es_index = ''
        self.__kind = ''
        self.__status = ''
        self.__pod_name = ''
        self.__vm_name = ''
        self.__scale = ''
        self.__data_dict = {}

    def save_error_logs(self):
        """
        This method uploads logs into elastic and s3 bucket in case of error
        @return:
        """
        if self._es_host:
            self.__data_dict['run_artifacts_url'] = os.path.join(self._run_artifacts_url,
                                                                f'{self._get_run_artifacts_hierarchy(workload_name=self._get_workload_file_name(self.__workload_name), is_file=True)}.tar.gz')
            self._upload_to_elasticsearch(index=self.__es_index, kind=self.__kind, status='failed',
                                          result=self.__data_dict)
            # verify that data upload to elastic search according to unique uuid
            self._verify_elasticsearch_data_uploaded(index=self.__es_index, uuid=self._uuid)

    def __create_vm_scale(self, vm_num: str):
        """
        This method creates vm in parallel
        """
        try:
            self._oc.create_async(yaml=os.path.join(f'{self._run_artifacts_path}', f'{self.__name}_{vm_num}.yaml'))
            self._oc.wait_for_vm_create(vm_name=f'{self.__vm_name}-{vm_num}')
        except Exception as err:
            # save run artifacts logs
            self.save_error_logs()
            raise err

    def __run_vm_scale(self, vm_num: str):
        """
        This method runs vm in parallel
        """
        try:
            self._oc.wait_for_ready(label=f'app=vdbench-{self._trunc_uuid}-{vm_num}', run_type='vm', label_uuid=False)
            # Create vm log should be direct after vm is ready
            self.__vm_name = self._create_vm_log(labels=[f'{self.__workload_name}-{self._trunc_uuid}-{vm_num}'])
            self.__status = self._oc.wait_for_vm_log_completed(vm_name=self.__vm_name, end_stamp=self.END_STAMP)
            self.__status = 'complete' if self.__status else 'failed'
            # prometheus queries
            self._prometheus_metrics_operation.finalize_prometheus()
            metric_results = self._prometheus_metrics_operation.run_prometheus_queries()
            prometheus_result = self._prometheus_metrics_operation.parse_prometheus_metrics(data=metric_results)
            # save run artifacts logs
            result_list = self._create_vm_run_artifacts(vm_name=f'{self.__workload_name}-{self._trunc_uuid}-{vm_num}', start_stamp=self.START_STAMP, end_stamp=self.END_STAMP, log_type='.csv')
            if self._es_host:
                # upload several run results
                for result in result_list:
                    result.update(prometheus_result)
                    self._upload_to_elasticsearch(index=self.__es_index, kind=self.__kind, status=self.__status, result=result)
                # verify that data upload to elastic search according to unique uuid
                self._verify_elasticsearch_data_uploaded(index=self.__es_index, uuid=self._uuid)
            self._oc.delete_vm_sync(
                yaml=os.path.join(f'{self._run_artifacts_path}', f'{self.__name}_{vm_num}.yaml'),
                vm_name=f'{self.__vm_name}-{vm_num}')
        except Exception as err:
            # save run artifacts logs
            self.save_error_logs()
            raise err

    def __delete_vm_scale(self, vm_num: str):
        """
        This method deletes vm in parallel
        """
        try:
            self._oc.delete_async(yaml=os.path.join(f'{self._run_artifacts_path}', f'{self.__name}_{vm_num}.yaml'))
        except Exception as err:
            # save run artifacts logs
            self.save_error_logs()
            raise err

    @logger_time_stamp
    def run(self):
        """
        This method runs the workload
        :return:
        """
        try:
            self._prometheus_metrics_operation.init_prometheus()
            self.__name = self._workload
            if self._run_type == 'test_ci':
                self.__es_index = 'vdbench-test-ci-results'
            else:
                self.__es_index = 'vdbench-results'
            self.__workload_name = self._workload.replace('_', '-')
            self.__vm_name = f'{self.__workload_name}-{self._trunc_uuid}'
            self.__kind = 'vm'
            self._environment_variables_dict['kind'] = 'vm'
            # create namespace
            self._oc.create_async(yaml=os.path.join(f'{self._run_artifacts_path}', 'namespace.yaml'))
            if not self._scale:
                self._oc.create_vm_sync(yaml=os.path.join(f'{self._run_artifacts_path}', f'{self.__name}.yaml'), vm_name=self.__vm_name)
                self._oc.wait_for_ready(label=f'app=vdbench-{self._trunc_uuid}', run_type='vm', label_uuid=False)
                # Create vm log should be direct after vm is ready
                self.__vm_name = self._create_vm_log(labels=[self.__workload_name])
                self.__status = self._oc.wait_for_vm_log_completed(vm_name=self.__vm_name, end_stamp=self.END_STAMP)
                self.__status = 'complete' if self.__status else 'failed'
                # save run artifacts logs
                result_list = self._create_vm_run_artifacts(vm_name=self.__vm_name, start_stamp=self.START_STAMP, end_stamp=self.END_STAMP, log_type='.csv')
                # prometheus queries
                self._prometheus_metrics_operation.finalize_prometheus()
                metric_results = self._prometheus_metrics_operation.run_prometheus_queries()
                prometheus_result = self._prometheus_metrics_operation.parse_prometheus_metrics(data=metric_results)
                if self._es_host:
                    # upload several run results
                    for result in result_list:
                        result.update(prometheus_result)
                        self._upload_to_elasticsearch(index=self.__es_index, kind=self.__kind, status=self.__status, result=result)
                    # verify that data upload to elastic search according to unique uuid
                    self._verify_elasticsearch_data_uploaded(index=self.__es_index, uuid=self._uuid)
                self._oc.delete_vm_sync(
                    yaml=os.path.join(f'{self._run_artifacts_path}', f'{self.__name}.yaml'),
                    vm_name=self.__vm_name)
            # scale
            else:
                self.__scale = int(self._scale)
                # create redis and state signals
                sync_pods = {'redis': 'redis', 'state_signals_exporter_pod': 'state-signals-exporter'}
                for pod, name in sync_pods.items():
                    if pod == 'redis':
                        pod_name = f'redis-master'
                    else:
                        pod_name = name
                    self._oc.create_pod_sync(yaml=os.path.join(f'{self._run_artifacts_path}', f'{pod}.yaml'), pod_name=pod_name)
                    self._oc.wait_for_initialized(label=f'app={name}', label_uuid=False)
                    self._oc.wait_for_ready(label=f'app={name}', label_uuid=False)
                # prepare scale run
                bulks = tuple(self.split_run_bulks(iterable=range(self._scale * len(self._scale_node_list)), limit=self._threads_limit))
                # create, run and delete vms
                for target in (self.__create_vm_scale, self.__run_vm_scale, self.__delete_vm_scale):
                    proc = []
                    for bulk in bulks:
                        for vm_num in bulk:
                            p = Process(target=target, args=(str(vm_num),))
                            p.start()
                            proc.append(p)
                        for p in proc:
                            p.join()
                        # sleep between bulks
                        time.sleep(self._bulk_sleep_time)
                        proc = []
                self._create_scale_logs()
                # delete redis and state signals
                for pod, name in sync_pods.items():
                    if pod == 'redis':
                        pod_name = f'redis-master'
                    else:
                        pod_name = name
                    self._oc.delete_pod_sync(yaml=os.path.join(f'{self._run_artifacts_path}', f'{pod}.yaml'), pod_name=pod_name)
            # delete namespace
            self._oc.delete_async(yaml=os.path.join(f'{self._run_artifacts_path}', 'namespace.yaml'))
        except ElasticSearchDataNotUploaded as err:
            self._oc.delete_vm_sync(
                yaml=os.path.join(f'{self._run_artifacts_path}', f'{self.__name}.yaml'),
                vm_name=self.__vm_name)
            raise err
        except Exception as err:
            # save run artifacts logs
            self.save_error_logs()
            raise err




import os

from benchmark_runner.common.oc.oc import OC
from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.common.ocp_resources.create_ocp_resource_operations import CreateOCPResourceOperations
from benchmark_runner.common.ocp_resources.create_ocp_resource_exceptions import ODFInstallationFailed


class CreateODF(CreateOCPResourceOperations):
    """
    This class is created ODF operator
    """
    ODF_CSV_NUM = 4

    def __init__(self, oc: OC, path: str, resource_list: list, worker_disk_ids: list, worker_disk_prefix: str):
        super().__init__(oc)
        self.__oc = oc
        self.__path = path
        self.__resource_list = resource_list
        self.__worker_disk_ids = worker_disk_ids
        self.__worker_disk_prefix = worker_disk_prefix

    @logger_time_stamp
    def create_odf(self, upgrade_version: str):
        """
        This method creates odf operator
        :param upgrade_version:
        :return:
        """
        if upgrade_version:
            self.__oc.apply_async(yaml=os.path.join(self.__path, '07_subscription.yaml'))
            logger.info(f'Wait till ODF upgrade to version: {upgrade_version}')
            self.verify_csv_installation(namespace='openshift-storage', operator='odf', upgrade_version=upgrade_version, csv_num=self.ODF_CSV_NUM)
        else:
            for resource in self.__resource_list:
                logger.info(f'run {resource}')
                if resource.endswith('.sh'):
                    # Ceph disk deletion - reference: https://rook.io/docs/rook/v1.12/Getting-Started/ceph-teardown/#delete-the-data-on-hosts
                    if '01_delete_disks.sh' == resource:
                        delete_node_disk = ''
                        for node, disk_ids in self.__worker_disk_ids.items():
                            for disk_id in disk_ids:
                                disk = f'/dev/disk/by-id/{self.__worker_disk_prefix}{disk_id}'
                                delete_node_disk += f"""sudo sgdisk --zap-all {disk}; sudo wipefs -a {disk}; sudo dd if=/dev/zero of='{disk}' bs=1M count=100 oflag=direct,dsync; sudo blkdiscard {disk}; sudo partprobe {disk};"""
                                logger.info(f'{node}: {delete_node_disk}')
                                self.__oc.run(cmd=f'chmod +x {os.path.join(self.__path, resource)}; {self.__path}/./{resource} "{node}" "{delete_node_disk}"')
                                delete_node_disk = ''
                    else:
                        self.__oc.run(cmd=f'chmod +x {os.path.join(self.__path, resource)}; {self.__path}/./{resource}')
                else:  # yaml
                    self.__oc.create_async(yaml=os.path.join(self.__path, resource))
                    if '04_local_volume_set.yaml' in resource:
                        # openshift local storage - diskmaker
                        self.wait_for_ocp_resource_create(operator='odf',
                                                          verify_cmd=r"""oc get pod -n openshift-local-storage -o jsonpath="{range .items[*]}{.metadata.name}{'\n'}{end}" | grep diskmaker | wc -l""",
                                                          count_disk_maker=True)
                        # openshift persistence volume - pv
                        self.wait_for_ocp_resource_create(operator='odf',
                                                          verify_cmd=r"""oc get pv -o jsonpath="{range .items[*]}{.metadata.name}{'\n'}{end}" | grep local | wc -l""",
                                                          count_openshift_storage=True)
                    if '07_subscription.yaml' in resource:
                        # Must be run after installing the storage cluster because CSVs sometimes fail
                        self.verify_csv_installation(namespace='openshift-storage', operator='odf', csv_num=self.ODF_CSV_NUM)
                    elif '08_storage_cluster.yaml' in resource:
                        self.wait_for_ocp_resource_create(operator='odf',
                                                          verify_cmd='oc get pod -n openshift-storage | grep osd | grep -v prepare | wc -l',
                                                          count_openshift_storage=True)
            # Verify ODF installation
            if not self.__oc.verify_odf_installation():
                raise ODFInstallationFailed
            return True




import time
from datetime import datetime
from typeguard import typechecked
import ast  # change string list to list
from enum import Enum

from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger, datetime_format
from benchmark_runner.common.clouds.IBM.ibm_exceptions import IBMMachineNotLoad
from benchmark_runner.common.clouds.BareMetal.bare_metal_operations import BareMetalOperations


class Actions(Enum):
    """
    IBM Actions
    """
    POWER_OFF = 'power-off'
    POWER_ON = 'power-on'
    REBOOT = 'reboot'


class IBMOperations(BareMetalOperations):
    """
    This class is responsible for all IBM cloud operations, all commands run on remote provision IBM host
    """
    SHORT_TIMEOUT = 600

    @typechecked
    def __init__(self, user: str):
        super().__init__(user)
        self._ibm_api_key = self._environment_variables_dict.get('ibm_api_key', '')
        self._ibm_worker_ids = self._environment_variables_dict.get('worker_ids', '')
        self._ibm_worker_ids_list = ast.literal_eval(self._ibm_worker_ids)

    def __ibm_login_cmd(self):
        """
        This method returns ibm login command
        :return:
        """
        return f'ibmcloud login --apikey {self._ibm_api_key} 1>/dev/null 2>&1'

    # private method: machine id
    def __get_ibm_machine_status(self, machine_id: str):
        """
        This method returns IBM machine status
        :return:
        """
        details = self._remote_ssh.run_command(command=f'ibmcloud sl hardware detail {machine_id}')
        item_results = details.split()
        for ind, item in enumerate(item_results):
            if item == 'Status':
                return item_results[ind+1]

    def __async_set_action_ibm_machine(self, action: str, machine_id: str):
        """
        This method reboots IBM machine id
        :param action:
        :param machine_id:
        :return:
        """
        self._remote_ssh.run_command(command=f'ibmcloud sl hardware {action} {machine_id} -f')

    def __wait_for_active_machine(self, machine_id: str, sleep_time=10, timeout=SHORT_TIMEOUT):
        """
        This method waits till IBM machine will be active
        :param machine_id:
        :return:
        """
        current_wait_time = 0
        while current_wait_time <= timeout:
            if self.__get_ibm_machine_status(machine_id=machine_id) == 'ACTIVE':
                return True
            # sleep for x seconds
            time.sleep(sleep_time)
            current_wait_time += sleep_time
        raise IBMMachineNotLoad()

    @staticmethod
    def __ibm_logout_cmd():
        """
        This method returns ibm logout command
        :return:
        """
        return 'ibmcloud logout'

    def run_ocp_installer(self):
        """
        This method run ocp assisted installer
        :return: True if installation success and raise exception if installation failed
        """
        self.update_provision_config()
        logger.info(f'Starting OCP assisted installer, Start time: {datetime.now().strftime(datetime_format)}')
        # Must add -t otherwise remote ssh of ansible will not end
        self._ssh.run(cmd=f"ssh -t provision \"{self.__ibm_login_cmd()};{self._install_ocp_cmd()}\" ")
        logger.info(f'OpenShift cluster {self._get_installation_version()} version is installed successfully, End time: {datetime.now().strftime(datetime_format)}')




# Custom exceptions must NOT expose data including hostnames, pathnames, IP addresses, etc. (security risk)
class SshError(Exception):
    """ Base class for all SSH error classes.
        All exceptions raised by the DSS library should inherit from this class. """
    pass


class RunCommandError(SshError):
    """
    This class for run command error
    """
    def __init__(self, err):
        self.message = f'Cannot run shell command: {err}'
        super(RunCommandError, self).__init__(self.message)


class SshConnectionError(SshError):
    """
    This class for connection error
    """
    def __init__(self):
        self.message = f'Missing connection credentials'
        super(SshConnectionError, self).__init__(self.message)


class SshConnectionTimedOut(SshError):
    """
    This class for connection timeout
    """
    def __init__(self, err):
        self.message = f'Timed Out : {err}'
        super(SshConnectionTimedOut, self).__init__(self.message)


class SshConnectionFailure(SshError):
    """
    This class for connection failure
    """
    def __init__(self, err):
        self.message = f'Server connection failure: {err}'
        super(SshConnectionFailure, self).__init__(self.message)


class PathNotExist(SshError):
    """
    This class error for folder that not exist
    """
    def __init__(self, err):
        self.message = f'{err}'
        super(PathNotExist, self).__init__(self.message)


class FileNotExist(SshError):
    """
    This class error for file that not exist
    """
    def __init__(self, err):
        self.message = f'{err}'
        super(PathNotExist, self).__init__(self.message)


class SFTPException(SshError):
    """
    This class through sftp exception
    """
    def __init__(self, err):
        self.message = {err}
        super(PathNotExist, self).__init__(self.message)


class IllegalFilename(SshError):
    """
    This class through Illegal file name error
    """
    def __init__(self, err):
        self.message = {err}
        super(PathNotExist, self).__init__(self.message)




import os
import time
import datetime
import argparse
from uuid import uuid4
from benchmark_runner.main.environment_variables_exceptions import ParseFailed


class EnvironmentVariables:
    """
    This class manage environment variable parameters
    """

    HAMMERDB_LSO_LEN = 4

    def __init__(self):

        self._environment_variables_dict = {}

        # env files override true ENV. Not best order, but easier to write :/
        # .env.generated can be auto-generated (by an external tool) based on the local cluster's configuration.

        for env in ".env", ".env.generated":
            try:
                with open(env) as f:
                    for line in f.readlines():
                        key, found, value = line.strip().partition("=")
                        if not found:
                            print("ERROR: invalid line in {env}: {line.strip()}")
                            continue
                        if key in os.environ:
                            continue  # prefer env to env file
                        os.environ[key] = value

            except FileNotFoundError:
                pass  # ignore

        ##################################################################################################
        # dynamic parameters - configure for local run
        # parameters for running workload

        # This path is GitHub actions runner path (benchmark-operator should be cloned here)
        self._environment_variables_dict['runner_path'] = EnvironmentVariables.get_env('RUNNER_PATH', '/tmp')
        # This path is for vm/pod/prometheus run artifacts
        self._environment_variables_dict['run_artifacts'] = EnvironmentVariables.get_env('RUN_ARTIFACTS', os.path.join(self._environment_variables_dict['runner_path'], 'benchmark-runner-run-artifacts'))
        # cluster: 'openshift'(Default)/ 'kubernetes'
        self._environment_variables_dict['cluster'] = EnvironmentVariables.get_env('CLUSTER', 'openshift')

        # dynamic parameters - configure for local run
        self._environment_variables_dict['workload'] = EnvironmentVariables.get_env('WORKLOAD', '')
        self._environment_variables_dict['kubeadmin_password'] = EnvironmentVariables.get_env('KUBEADMIN_PASSWORD', '')

        # PIN=node selector
        self._environment_variables_dict['pin_node_benchmark_operator'] = EnvironmentVariables.get_env('PIN_NODE_BENCHMARK_OPERATOR', '')
        self._environment_variables_dict['pin_node1'] = EnvironmentVariables.get_env('PIN_NODE1', '')
        self._environment_variables_dict['pin_node2'] = EnvironmentVariables.get_env('PIN_NODE2', '')

        # ElasticSearch
        self._environment_variables_dict['elasticsearch'] = EnvironmentVariables.get_env('ELASTICSEARCH', '')
        self._environment_variables_dict['elasticsearch_port'] = EnvironmentVariables.get_env('ELASTICSEARCH_PORT', '')
        self._environment_variables_dict['elasticsearch_user'] = EnvironmentVariables.get_env('ELASTICSEARCH_USER', '')
        self._environment_variables_dict['elasticsearch_password'] = EnvironmentVariables.get_env('ELASTICSEARCH_PASSWORD', '')
        # 'http'(Default) / 'https' to use SSL to connect ElasticSearch
        self._environment_variables_dict['elasticsearch_url_protocol'] = EnvironmentVariables.get_env('ELASTICSEARCH_URL_PROTOCOL', 'http')

        # Workaround for Kata CPU offline problem in 4.9/4.10
        self._environment_variables_dict['kata_cpuoffline_workaround'] = EnvironmentVariables.get_boolean_from_environment('KATA_CPUOFFLINE_WORKAROUND', False)
        # Kata thread-pool-size, default 16
        self._environment_variables_dict['kata_thread_pool_size'] = EnvironmentVariables.get_boolean_from_environment('KATA_THREAD_POOL_SIZE', '16')

        # Scale Per Node
        self._environment_variables_dict['scale'] = EnvironmentVariables.get_env('SCALE', '')
        # list of nodes per pod/vm, scale number per node, e.g: [ 'master-1', 'master-2' ] - run 1 pod/vm in each node
        self._environment_variables_dict['scale_nodes'] = EnvironmentVariables.get_env('SCALE_NODES', "")
        self._environment_variables_dict['bulk_sleep_time'] = EnvironmentVariables.get_env('BULK_SLEEP_TIME', '30')
        # CPU processors = threads limit
        self._environment_variables_dict['threads_limit'] = EnvironmentVariables.get_env('THREADS_LIMIT', '')
        # redis for synchronization
        self._environment_variables_dict['redis'] = EnvironmentVariables.get_env('REDIS', '')

        # prometheus snap interval
        self._environment_variables_dict['prometheus_snap_interval'] = EnvironmentVariables.get_env('PROMETHEUS_SNAP_INTERVAL', '30')

        # windows url
        self._environment_variables_dict['windows_url'] = EnvironmentVariables.get_env('WINDOWS_URL', '')
        # Delete all resources before and after the run, default True
        self._environment_variables_dict['delete_all'] = EnvironmentVariables.get_boolean_from_environment('DELETE_ALL', True)
        # Verification only, without running or deleting any resources, default False
        self._environment_variables_dict['verification_only'] = EnvironmentVariables.get_boolean_from_environment('VERIFICATION_ONLY', False)

        # default parameter - change only if needed
        # Parameters below related to 'run_workload()'
        self._environment_variables_dict['workloads'] = ['stressng_pod', 'stressng_vm', 'stressng_kata',
                                                         'uperf_pod', 'uperf_vm', 'uperf_kata',
                                                         'hammerdb_pod_mariadb', 'hammerdb_vm_mariadb', 'hammerdb_kata_mariadb',
                                                         'hammerdb_pod_mariadb_lso', 'hammerdb_vm_mariadb_lso', 'hammerdb_kata_mariadb_lso',
                                                         'hammerdb_pod_postgres', 'hammerdb_vm_postgres', 'hammerdb_kata_postgres',
                                                         'hammerdb_pod_postgres_lso', 'hammerdb_vm_postgres_lso', 'hammerdb_kata_postgres_lso',
                                                         'hammerdb_pod_mssql', 'hammerdb_vm_mssql', 'hammerdb_kata_mssql',
                                                         'hammerdb_pod_mssql_lso', 'hammerdb_vm_mssql_lso', 'hammerdb_kata_mssql_lso',
                                                         'vdbench_pod', 'vdbench_kata', 'vdbench_vm',
                                                         'clusterbuster', 'bootstorm_vm', 'windows_vm']
        # Workloads namespaces
        self._environment_variables_dict['workload_namespaces'] = {
            'stressng': 'benchmark-operator',
            'hammerdb': 'benchmark-operator',
            'uperf': 'benchmark-operator',
            'vdbench': 'benchmark-runner',
            'clusterbuster': 'clusterbuster',
            'bootstorm': 'benchmark-runner',
            'windows': 'benchmark-runner'
        }

        # Update namespace
        base_workload = self._environment_variables_dict['workload'].split('_')[0]
        if EnvironmentVariables.get_env('NAMESPACE'):
            self._environment_variables_dict['namespace'] = EnvironmentVariables.get_env('NAMESPACE')
        elif base_workload in self._environment_variables_dict['workload_namespaces']:
            default_namespace = self._environment_variables_dict['workload_namespaces'][base_workload]
            self._environment_variables_dict['namespace'] = EnvironmentVariables.get_env('NAMESPACE', default_namespace)
        else:
            # TBD if this is not set
            self._environment_variables_dict['namespace'] = 'benchmark-operator'

        # run workload with odf pvc True/False. True=ODF(default), False=Ephemeral
        self._environment_variables_dict['odf_pvc'] = EnvironmentVariables.get_boolean_from_environment('ODF_PVC', True)
        if base_workload == 'hammerdb':
            if len(self._environment_variables_dict['workload'].split('_')) == self.HAMMERDB_LSO_LEN:
                self._environment_variables_dict['storage_type'] = self._environment_variables_dict['workload'].split('_')[self.HAMMERDB_LSO_LEN-1]
            elif self._environment_variables_dict['odf_pvc']:
                self._environment_variables_dict['storage_type'] = 'odf'
            else:
                self._environment_variables_dict['storage_type'] = 'ephemeral'

        # LSO Disk id - auto-detect when located on worker-2
        self._environment_variables_dict['lso_disk_id'] = EnvironmentVariables.get_env('LSO_DISK_ID', '')
        self._environment_variables_dict['lso_node'] = EnvironmentVariables.get_env('LSO_NODE', '')
        # Workloads that required ODF
        self._environment_variables_dict['workloads_odf_pvc'] = ['vdbench', 'hammerdb']
        # This parameter get from Test_CI.yml file
        self._environment_variables_dict['build_version'] = EnvironmentVariables.get_env('BUILD_VERSION', '1.0.0')
        # collect system metrics True/False - required by benchmark-operator: @todo disable in OCP4.12.0-rc.5 - pod not create
        self._environment_variables_dict['system_metrics'] = EnvironmentVariables.get_boolean_from_environment('SYSTEM_METRICS', False)
        # CI status update once at the end of CI pass/failed
        self._environment_variables_dict['ci_status'] = EnvironmentVariables.get_env('CI_STATUS', '')
        # Valid run types
        self._environment_variables_dict['run_types'] = ['test_ci', 'func_ci', 'perf_ci', 'release']
        # Run type test_ci/func_ci/perf_ci, default test_ci same environment as func_ci
        self._environment_variables_dict['run_type'] = EnvironmentVariables.get_env('RUN_TYPE', 'test_ci')
        self._environment_variables_dict['runner_type'] = EnvironmentVariables.get_env('RUNNER_TYPE')
        self._environment_variables_dict['config_from_args'] = EnvironmentVariables.get_boolean_from_environment('CONFIG_FROM_ARGS', False)
        self._environment_variables_dict['template_in_workload_dir'] = EnvironmentVariables.get_env('TEMPLATE_IN_WORKLOAD_DIR')

        # Run uuid
        self._environment_variables_dict['uuid'] = EnvironmentVariables.get_env('UUID', str(uuid4()))
        self._environment_variables_dict['trunc_uuid'] = self._environment_variables_dict['uuid'].split('-')[0]
        # Benchmark runner IBM Cloud Object Storage run artifacts hierarchy, not part of a POSIX path ('/' a key seperator, '-' file name convenstion )
        self._environment_variables_dict['date_key'] = datetime.datetime.now().strftime("%Y/%m/%d")
        self._environment_variables_dict['time_stamp_format'] = os.path.join(datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d-%H-%M-%S'))
        # Benchmark runner local run artifacts path with time stamp format

        self._environment_variables_dict['run_artifacts_path'] = EnvironmentVariables.get_env('RUN_ARTIFACTS_PATH')
        if not self._environment_variables_dict['run_artifacts_path']:
            self._environment_variables_dict['run_artifacts_path'] = os.path.join(self._environment_variables_dict['run_artifacts'], f"{self._environment_variables_dict['workload'].replace('_', '-')}-{self._environment_variables_dict['time_stamp_format']}")

        # True/False: default False
        self._environment_variables_dict['save_artifacts_local'] = EnvironmentVariables.get_boolean_from_environment('SAVE_ARTIFACTS_LOCAL', False)
        # True/False: default False
        self._environment_variables_dict['enable_prometheus_snapshot'] = EnvironmentVariables.get_boolean_from_environment('ENABLE_PROMETHEUS_SNAPSHOT', False)
        # end dynamic parameters - configure for local run
        ##################################################################################################

        # ** DO NOT CHANGE THE PARAMETERS BELOW **
        # Constant parameters

        # Azure Parameters
        # Cluster Operations: START/ STOP
        self._environment_variables_dict['azure_cluster_operation'] = EnvironmentVariables.get_env('AZURE_CLUSTER_OPERATION', '')
        self._environment_variables_dict['azure_clientid'] = EnvironmentVariables.get_env('AZURE_CLIENTID', '')
        self._environment_variables_dict['azure_secret'] = EnvironmentVariables.get_env('AZURE_SECRET', '')
        self._environment_variables_dict['azure_tenantid'] = EnvironmentVariables.get_env('AZURE_TENANTID', '')
        self._environment_variables_dict['azure_subscriptionid'] = EnvironmentVariables.get_env('AZURE_SUBSCRIPTIONID', '')
        self._environment_variables_dict['azure_resource_group_name'] = EnvironmentVariables.get_env('AZURE_RESOURCE_GROUP_NAME', '')
        self._environment_variables_dict['azure_vm_name'] = EnvironmentVariables.get_env('AZURE_VM_NAME', '')

        # CLUSTRBUSTER data
        self._environment_variables_dict['clusterbuster_workload'] = EnvironmentVariables.get_env('CLUSTERBUSTER_WORKLOAD', '')
        self._environment_variables_dict['clusterbuster_uuid'] = EnvironmentVariables.get_env('CLUSTERBUSTER_UUID', '')

        # IBM data
        self._environment_variables_dict['region_name'] = EnvironmentVariables.get_env('IBM_REGION_NAME', '')
        # None(default) - must for unittest
        self._environment_variables_dict['endpoint_url'] = EnvironmentVariables.get_env('IBM_ENDPOINT_URL', None)
        self._environment_variables_dict['access_key_id'] = EnvironmentVariables.get_env('IBM_ACCESS_KEY_ID', '')
        self._environment_variables_dict['secret_access_key'] = EnvironmentVariables.get_env('IBM_SECRET_ACCESS_KEY', '')
        self._environment_variables_dict['bucket'] = EnvironmentVariables.get_env('IBM_BUCKET', '')
        self._environment_variables_dict['key'] = EnvironmentVariables.get_env('IBM_KEY', '')

        # Grafana
        self._environment_variables_dict['grafana_url'] = EnvironmentVariables.get_env('GRAFANA_URL', '')
        self._environment_variables_dict['grafana_api_key'] = EnvironmentVariables.get_env('GRAFANA_API_KEY', '')
        self._environment_variables_dict['grafana_json_path'] = EnvironmentVariables.get_env('GRAFANA_JSON_PATH', '')
        self._environment_variables_dict['main_libsonnet_path'] = EnvironmentVariables.get_env('MAIN_LIBSONNET_PATH', '')
        self._environment_variables_dict['grafana_folder_name'] = EnvironmentVariables.get_env('GRAFANA_FOLDER_NAME', '')

        # Parameters below related to 'install_ocp()'
        # MANDATORY for OCP install: assisted installer version i.e. 'latest-4.16' or 'latest-4.16.0-rc' or '4.16.0'. Verify the version at https://openshift-release.apps.ci.l2s4.p1.openshiftapps.com/
        self._environment_variables_dict['install_ocp_version'] = EnvironmentVariables.get_env('INSTALL_OCP_VERSION', '')
        # There are 4 options: run_bare_metal_ocp_installer/ verify_bare_metal_install_complete/ run_ibm_ocp_installer/ verify_ibm_install_complete
        self._environment_variables_dict['install_step'] = EnvironmentVariables.get_env('INSTALL_STEP', '')

        # MANDATORY for OCP upgrade: Specify the version as '4.15.22'. Ensure that the upgrade version is stable by checking: https://github.com/openshift/cincinnati-graph-data/tree/master/channels )
        self._environment_variables_dict['upgrade_ocp_version'] = EnvironmentVariables.get_env('UPGRADE_OCP_VERSION','')
        # There are 2 options: run_bare_metal_ocp_upgrade/ verify_bare_metal_upgrade_complete
        self._environment_variables_dict['upgrade_step'] = EnvironmentVariables.get_env('UPGRADE_STEP', '')

        # SNO or empty for regular
        self._environment_variables_dict['cluster_type'] = EnvironmentVariables.get_env('CLUSTER_TYPE', '')
        # For SNO: choose 1 master, dictionary: {'master': ['master-0', 'master-1', 'master-2'], 'worker': ['worker-0', 'worker-1', 'worker-2' ] }
        self._environment_variables_dict['expected_nodes'] = EnvironmentVariables.get_env('EXPECTED_NODES', "")
        # GitHub repository - for credentials updating
        self._environment_variables_dict['github_repository_short'] = EnvironmentVariables.get_env('GITHUB_REPOSITORY_SHORT', '')

        # Parameters below related to 'install_resource()'
        # MANDATORY for OCP resource install: True for install resources, default False
        self._environment_variables_dict['install_ocp_resources'] = EnvironmentVariables.get_boolean_from_environment('INSTALL_OCP_RESOURCES', False)
        # cnv version
        self._environment_variables_dict['cnv_version'] = EnvironmentVariables.get_env('CNV_VERSION', '')
        # cnv nightly channel, True: nightly / False: stable, default True
        self._environment_variables_dict['cnv_nightly_channel'] = EnvironmentVariables.get_boolean_from_environment('CNV_NIGHTLY_CHANNEL', True)
        # QUAY_USERNAME for nightly build
        self._environment_variables_dict['quay_username'] = EnvironmentVariables.get_env('QUAY_USERNAME', '')
        # QUAY_PASSWORD for nightly build
        self._environment_variables_dict['quay_password'] = EnvironmentVariables.get_env('QUAY_PASSWORD', '')
        # lso version
        self._environment_variables_dict['lso_version'] = EnvironmentVariables.get_env('LSO_VERSION', '')
        # odf version
        self._environment_variables_dict['odf_version'] = EnvironmentVariables.get_env('ODF_VERSION', '')
        # custom kata version, if empty fetch auto latest version
        self._environment_variables_dict['kata_csv'] = EnvironmentVariables.get_env('KATA_CSV', '')
        # number of odf disk for discovery
        self._environment_variables_dict['num_odf_disk'] = EnvironmentVariables.get_env('NUM_ODF_DISK', 4)
        self._environment_variables_dict['worker_disk_ids'] = EnvironmentVariables.get_env('WORKER_DISK_IDS', "")
        self._environment_variables_dict['worker_disk_prefix'] = EnvironmentVariables.get_env('WORKER_DISK_PREFIX', '')
        # install resources list "[ 'lso', 'odf', 'cnv', 'infra', 'custom' ]"
        self._environment_variables_dict['install_resources_list'] = EnvironmentVariables.get_env('INSTALL_RESOURCES_LIST', "")

        # Parameters below related to 'install_ocp()' and 'install_resource()'
        # Mandatory: OCP environment flavor PERF or FUNC for updating GitHub secrets
        self._environment_variables_dict['ocp_env_flavor'] = EnvironmentVariables.get_env('OCP_ENV_FLAVOR', 'PERF')
        # IBM data
        self._environment_variables_dict['ibm_api_key'] = EnvironmentVariables.get_env('IBM_API_KEY', '')
        # GitHub token
        self._environment_variables_dict['github_token'] = EnvironmentVariables.get_env('GITHUB_TOKEN', '')
        self._environment_variables_dict['worker_ids'] = EnvironmentVariables.get_env(f'WORKER_IDS', "")
        self._environment_variables_dict['provision_ip'] = EnvironmentVariables.get_env(f'PROVISION_IP', '')
        # Placed on secret only
        self._environment_variables_dict['provision_private_key'] = EnvironmentVariables.get_env(f'PROVISION_PRIVATE_KEY', '')
        # For internal private key path
        self._environment_variables_dict['provision_private_key_path'] = EnvironmentVariables.get_env(f'PROVISION_PRIVATE_KEY_PATH', '')
        self._environment_variables_dict['container_private_key_path'] = EnvironmentVariables.get_env('CONTAINER_PRIVATE_KEY_PATH', '')
        self._environment_variables_dict['provision_user'] = EnvironmentVariables.get_env(f'PROVISION_USER', '')
        self._environment_variables_dict['provision_port'] = EnvironmentVariables.get_env(f'PROVISION_PORT', '')
        self._environment_variables_dict['provision_kubeadmin_password_path'] = EnvironmentVariables.get_env(f'PROVISION_KUBEADMIN_PASSWORD_PATH', '')
        self._environment_variables_dict['provision_kubeconfig_path'] = EnvironmentVariables.get_env(f'PROVISION_KUBECONFIG_PATH', '')
        self._environment_variables_dict['container_kubeconfig_path'] = EnvironmentVariables.get_env('CONTAINER_KUBECONFIG_PATH', '')
        self._environment_variables_dict['provision_installer_path'] = EnvironmentVariables.get_env(f'PROVISION_INSTALLER_PATH', '')
        self._environment_variables_dict['provision_installer_cmd'] = EnvironmentVariables.get_env(f'PROVISION_INSTALLER_CMD', '')
        self._environment_variables_dict['provision_installer_log'] = EnvironmentVariables.get_env(f'PROVISION_INSTALLER_LOG', '')

        # timeout 0<=: forever, >0: second (installer)
        self._environment_variables_dict['provision_timeout'] = EnvironmentVariables.get_env(f'PROVISION_TIMEOUT', '10800')
        # timeout 0<=: forever, >0: second
        self._environment_variables_dict['timeout'] = EnvironmentVariables.get_env(f'TIMEOUT', '3600')

        # Benchmark runner run artifacts url
        self._environment_variables_dict['run_artifacts_url'] = EnvironmentVariables.get_env(f'RUN_ARTIFACTS_URL', '')

        # Parameters below related to 'update_ci_status()' - No need to configure update auto by ci
        # CI run time
        self._environment_variables_dict['ci_minutes_time'] = EnvironmentVariables.get_env('CI_MINUTES_TIME', 0)
        # Get this parameter from install resource process
        self._environment_variables_dict['ocp_resource_install_minutes_time'] = EnvironmentVariables.get_env('OCP_RESOURCE_INSTALL_MINUTES_TIME', 0)
        # benchmark-runner last commit id
        self._environment_variables_dict['benchmark_runner_id'] = EnvironmentVariables.get_env('BENCHMARK_RUNNER_ID', '')
        # benchmark-operator last commit id
        self._environment_variables_dict['benchmark_operator_id'] = EnvironmentVariables.get_env('BENCHMARK_OPERATOR_ID', '')
        # benchmark-wrapper last commit id
        self._environment_variables_dict['benchmark_wrapper_id'] = EnvironmentVariables.get_env('BENCHMARK_WRAPPER_ID', '')

        # Node Selector functionality
        self._environment_variables_dict['pin'] = bool(self._environment_variables_dict['pin_node1'])
        # if pin_node2 not exist, get pin_node1 value
        if self._environment_variables_dict['pin_node1'] and not self._environment_variables_dict['pin_node2']:
            self._environment_variables_dict['pin_node2'] = self._environment_variables_dict['pin_node1']

        # ElasticSearch url
        if self._environment_variables_dict.get('elasticsearch_password', ''):
            self._environment_variables_dict['elasticsearch_url'] = f"{self._environment_variables_dict['elasticsearch_url_protocol']}://{self._environment_variables_dict.get('elasticsearch_user', '')}:{self._environment_variables_dict.get('elasticsearch_password', '')}@{self._environment_variables_dict.get('elasticsearch', '')}:{self._environment_variables_dict.get('elasticsearch_port', '')}"
        else:
            if self._environment_variables_dict['elasticsearch'] and self._environment_variables_dict.get('elasticsearch_port', ''):
                self._environment_variables_dict['elasticsearch_url'] = f"{self._environment_variables_dict['elasticsearch_url_protocol']}://{self._environment_variables_dict.get('elasticsearch', '')}:{self._environment_variables_dict.get('elasticsearch_port', '')}"
            else:
                self._environment_variables_dict['elasticsearch_url'] = ''

        # OpenShift or kubernetes support, OpenShift: oc, kubectl || kubernetes: kubectl
        if self._environment_variables_dict['cluster'] == 'kubernetes':
            self._environment_variables_dict['cli'] = 'kubectl'
            self._environment_variables_dict['odf_pvc'] = False
            self._environment_variables_dict['enable_prometheus_snapshot'] = None
        else:
            self._environment_variables_dict['cli'] = EnvironmentVariables.get_env('CLI', 'oc')

    @property
    def workloads_list(self):
        """
        This method is getter
        """
        return self._environment_variables_dict['workloads']

    @property
    def environment_variables_dict(self):
        """
        This method is getter
        """
        return self._environment_variables_dict

    @property
    def run_types_list(self):
        """
        This method is getter
        """
        return self._environment_variables_dict['run_types']

    @environment_variables_dict.setter
    def environment_variables_dict(self, value: dict):
        """
        This method is setter
        """
        self._environment_variables_dict = value

    def get_workload_namespace(self, workload: str):
        """
        Return the workload namespace for a given workload
        """
        if workload in self._environment_variables_dict['workloads'] and workload.split('_')[0] in self._environment_variables_dict['workload_namespaces']:
            return self._environment_variables_dict['workload_namespaces'][workload.split('_')[0]]
        else:
            return None

    @staticmethod
    def to_bool(arg, def_val: bool = None):
        if isinstance(arg, bool):
            return arg
        if isinstance(arg, (int, float)):
            return arg != 0
        if isinstance(arg, str):
            arg = arg.lower()
            if arg == 'true' or arg == 'yes':
                return True
            elif arg == 'false' or arg == 'no':
                return False
            try:
                arg1 = int(arg)
                return arg1 != 0
            except Exception:
                pass
        if def_val is not None:
            return def_val
        raise ParseFailed(f'Cannot parse {arg} as a boolean value')

    @staticmethod
    def get_env(var: str, defval=''):
        lcvar = var.lower()
        dashvar = lcvar.replace('_', '-')
        parser = argparse.ArgumentParser(description = 'Run BenchmarkRunner', allow_abbrev=False)
        if lcvar == dashvar:
            parser.add_argument(f"--{lcvar}", default=os.environ.get(var, defval), type=str, metavar='String', help=var)
        else:
            parser.add_argument(f"--{lcvar}", f"--{dashvar}", default=os.environ.get(var, defval), type=str, metavar='String', help=var)
        args, ignore = parser.parse_known_args()
        if hasattr(args, lcvar):
            return getattr(args, lcvar)
        else:
            return os.environ.get(var, defval)

    @staticmethod
    def get_boolean_from_environment(var: str, defval: bool):
        return EnvironmentVariables.to_bool(EnvironmentVariables.get_env(var), defval)


environment_variables = EnvironmentVariables()




class BareMetalError(Exception):
    """ Base class for all bare metal error classes.
    All exceptions raised by the Bare Metal library should inherit from this class. """
    pass


class MissingMasterNodes(BareMetalError):
    """
    This exception class is for missing master nodes
    """
    def __init__(self, missing_node: set):
        self.message = f'Missing master nodes: {missing_node}'
        super(MissingMasterNodes, self).__init__(self.message)


class MissingWorkerNodes(BareMetalError):
    """
    This exception class is for missing worker nodes.
    """
    def __init__(self, missing_node: set):
        self.message = f'Missing worker nodes: {missing_node}'
        super(MissingWorkerNodes, self).__init__(self.message)


class OCPInstallationFailed(BareMetalError):
    """
    This exception class indicates that the Bare-Metal OpenShift installation has failed
    """
    def __init__(self, logs: str):
        self.message = (f'Bare-metal OpenShift Container Platform (OCP) installation failed. '
                        f'OpenShift installer logs details: \n {logs}')
        super(OCPInstallationFailed, self).__init__(self.message)


class OCPUpgradeFailed(BareMetalError):
    """
    This exception class indicates that the Bare-Metal OpenShift upgrade has failed
    """
    def __init__(self, status: str):
        self.message = (f'Bare-metal OpenShift Container Platform (OCP) upgrade failed. '
                        f'Upgrade status: {status}')
        super(OCPUpgradeFailed, self).__init__(self.message)




import inspect
import importlib

from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.workloads.workloads_operations import WorkloadsOperations


class Workloads(WorkloadsOperations):
    """
    This class run workload
    """
    def __init__(self):
        """
        All inherit from WorkloadsOperations
        """
        super().__init__()

    @logger_time_stamp
    def run(self):
        """
        The method run workload
        :return:
        """

        # kata use pod module - replace kata to pod
        workload = self._workload.replace('kata', 'pod')
        # load the workload module before doing anything else (in case it fails)
        workload_module = importlib.import_module(f'benchmark_runner.workloads.{workload}')

        try:
            if not self._verification_only:
                self.initialize_workload()
            success = True
            # extract workload module and class
            for cls in inspect.getmembers(workload_module, inspect.isclass):
                if workload.replace('_', '').lower() == cls[0].lower():
                    if cls[1]().run() == False:
                        success = False
            if not self._verification_only:
                self.finalize_workload()
            return success

        except Exception as err:
            self.finalize_workload()
            raise err




import os
import time
from typeguard import typechecked

from benchmark_runner.common.oc.oc import OC
from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.common.ocp_resources.create_ocp_resource_exceptions import KataInstallationFailed
from benchmark_runner.common.ocp_resources.create_ocp_resource_operations import CreateOCPResourceOperations
from benchmark_runner.main.environment_variables import environment_variables


class CreateKata(CreateOCPResourceOperations):
    """
    This class is created Kata operator
    """
    def __init__(self, oc: OC, path: str, resource_list: list):
        super().__init__(oc)
        self.__oc = oc
        self.__path = path
        self.__resource_list = resource_list

    def __install_and_wait_for_resource_with_retry(self, yaml_file: str, resource_type: str, resource: str):
        """
        This method creates a resource where the creation process itself may fail
        without indicating an error and then have to be retried.
        :param yaml_file:YAML file to create the resource
        :param resource_type:type of resource to create
        :param resource: name of resource to create
        :return:
        """
        timeout = int(environment_variables.environment_variables_dict['timeout'])
        current_wait_time = 0
        while timeout <= 0 or current_wait_time < timeout:
            self.__oc.create_async(yaml_file)
            # We cannot wait for a condition here, because the
            # create_async may simply not work even if it returns success.
            time.sleep(OC.SLEEP_TIME)
            if self.__oc.run(f'if oc get {resource_type} {resource} > /dev/null 2>&1 ; then echo succeeded; fi') == 'succeeded':
                return True
            current_wait_time += OC.SLEEP_TIME
        return False

    @logger_time_stamp
    def create_kata(self):
        """
        This method creates kata resource
        :return:
        """
        for resource in self.__resource_list:
            logger.info(f'run {resource}')
            if '01_operator.yaml' == resource:
                # Wait for kataconfig CRD to exist
                self.__oc.create_async(os.path.join(self.__path, resource))
                self.wait_for_ocp_resource_create(operator='kata',
                                                  verify_cmd='if oc get crd kataconfigs.kataconfiguration.openshift.io >/dev/null 2>&1 ; then echo succeeded ; fi',
                                                  status='succeeded')
            elif '02_config.yaml' == resource:
                # This one's tricky.  The problem is that it appears
                # that the kataconfigs CRD can exist, but attempting
                # to apply it doesn't "take" unless some other things
                # are already up.  So we have to keep applying the
                # kataconfig until it's present.
                if not self.__install_and_wait_for_resource_with_retry(os.path.join(self.__path, resource), 'kataconfig', 'example-kataconfig'):
                    raise KataInstallationFailed('Failed to apply kataconfig resource')
                # Next, we have to wait for the kata bits to actually install
                self.wait_for_ocp_resource_create(operator='kata',
                                                  verify_cmd="""oc get kataconfig -o jsonpath='{.items[0].status.conditions[?(@.type == "InProgress")].status}'""",
                                                  status='False')
            elif resource.endswith('.sh'):
                self.__oc.run(cmd=f'chmod +x {os.path.join(self.__path, resource)}; {os.path.join(self.__path, resource)}')
        return True




import os
from jinja2 import Template
from benchmark_runner.main.environment_variables import environment_variables
from benchmark_runner.common.oc.oc import OC


def render_yaml_string(template_str: str, environment_variable_dict: dict = None):
    """
    This method renders a YAML string from the environment or alternative optional environment
    :param dir_path:
    :param yaml_file:
    :param environment_variable_dict: add it for testing purpose
    :return yaml text:
    """
    if not environment_variable_dict:
        environment_variable_dict = environment_variables.environment_variables_dict
    # Allow faking the Prometheus token so that we can generate deterministic test files
    if 'prom_token_override' in environment_variable_dict:
        environment_variable_dict['prom_token'] = environment_variable_dict.get('prom_token_override', '')
    else:
        oc = OC(kubeadmin_password=environment_variable_dict.get('kubeadmin_password', ''))
        # set prom token, skip if running on kubernetes just for now
        if environment_variable_dict.get("cluster") != 'kubernetes':
            environment_variable_dict['prom_token'] = oc.get_prom_token()
    return Template(template_str, keep_trailing_newline=True).render(environment_variable_dict)


def render_yaml_file(dir_path: str, yaml_file: str, environment_variable_dict: dict = None):
    """
    This method renders a YAML file from the environment or alternative optional environment
    :param dir_path:
    :param yaml_file:
    :param environment_variable_dict: Optional environment dictionary, default the Linux environment
    :return yaml text:
    """
    # Jinja render yaml file
    with open(os.path.join(dir_path, yaml_file)) as f:
        return render_yaml_string(f.read(), environment_variable_dict)



import ipywidgets as widgets
from datetime import datetime, time
from IPython.display import display
from benchmark_runner.common.elasticsearch.elasticsearch_operations import ElasticSearchOperations


class ElasticSearchWidgets:
    """
    This class contains elasticsearch widgets
    """

    DEFAULT_TIME = 0

    def __init__(self, elasticsearch: ElasticSearchOperations):
        # Need elasticsearch object to get indexes
        self.__elasticsearch = elasticsearch

    @staticmethod
    def get_start_datetime():
        """
        This method return changes global start_datetime
        @return:
        """
        return start_datetime

    @staticmethod
    def get_end_datetime():
        """
        This method return changes global end_datetime
        @return:
        """
        return end_datetime

    @staticmethod
    def get_elastic_index():
        """
        This method return changes global elastic_index
        @return:
        """
        return elastic_index

    def get_perfci_indexes(self):
        """
        This method returns sorted perfci elasticsearch indexes that contains 'results' or 'status'
        @return:
        """
        # Use list comprehension to filter strings containing 'result' or 'status'
        return [item for item in self.__elasticsearch.get_all_indexes() if 'result' in item or 'status' in item]

    def index_dropdown(self):
        """
        index dropdown widget
        @return:
        """
        dropdown = widgets.Dropdown(
            options=self.get_perfci_indexes(),
            value=self.get_perfci_indexes()[0],
            description='Choose INDEX & DATES:',
            style={'description_width': 'initial'},
            layout=widgets.Layout(width='500px')
        )
        elastic_index = dropdown.value

        def on_dropdown_change(change):
            global elastic_index
            elastic_index = change.new

        dropdown.observe(on_dropdown_change, names='value')
        display(dropdown)
        return elastic_index

    def choose_datetime(self):
        """
        datetime widget
        @return:
        """
        start_datetime = None
        end_datetime = None
        start_date_picker = widgets.DatePicker(description='Start Date')
        end_date_picker = widgets.DatePicker(description='End Date')
        start_time_picker = widgets.TimePicker(description='Start Time', value=time(hour=self.DEFAULT_TIME, minute=self.DEFAULT_TIME, second=self.DEFAULT_TIME))
        end_time_picker = widgets.TimePicker(description='End Time', value=time(hour=self.DEFAULT_TIME, minute=self.DEFAULT_TIME, second=self.DEFAULT_TIME))

        def on_value_change(change):
            global start_datetime
            global end_datetime
            start_date = start_date_picker.value
            end_date = end_date_picker.value
            start_time = start_time_picker.value or time(hour=0, minute=0, second=0)
            end_time = end_time_picker.value or time(hour=0, minute=0, second=0)
            start_datetime = datetime.combine(start_date, start_time)
            end_datetime = datetime.combine(end_date, end_time)

        start_date_picker.observe(on_value_change, names='value')
        end_date_picker.observe(on_value_change, names='value')
        start_time_picker.observe(on_value_change, names='value')
        end_time_picker.observe(on_value_change, names='value')

        display(start_date_picker)
        display(start_time_picker)
        display(end_date_picker)
        display(end_time_picker)



import paramiko
import os
import socket

from benchmark_runner.common.remote_ssh.connection_data import ConnectionData
from benchmark_runner.common.remote_ssh.remote_ssh_exceptions import SshConnectionError, SshConnectionFailure, \
    RunCommandError, \
    SshConnectionTimedOut, PathNotExist, FileNotExist, SFTPException, IllegalFilename


class RemoteSsh:
    """
    This class is running remote ssh commands
    """

    def __init__(self, connection_data: ConnectionData):
        self.__p_client = paramiko.SSHClient()
        self.__p_client.set_missing_host_key_policy(
            paramiko.AutoAddPolicy())  # This script doesn't work for me unless this line is added!

        self.__p_sftp = None
        # remote host connection data
        self.__hostname = connection_data.host_name
        self.__port = connection_data.port
        self.__username = connection_data.user_name
        self.__key_file = connection_data.ssh_key
        self.__password = connection_data.password
        self.__cmd_timeout = connection_data.timeout

    def connect(self):
        """ This method connect to remote machine """
        try:
            # For connection with ssh key
            if self.__key_file:
                key = paramiko.RSAKey.from_private_key_file(self.__key_file)
                self.__p_client.connect(hostname=self.__hostname, port=self.__port, username=self.__username,
                                        pkey=key, timeout=self.__cmd_timeout)
            # For password
            elif self.__password:
                self.__p_client.connect(hostname=self.__hostname, port=self.__port, username=self.__username,
                                        password=self.__password, timeout=self.__cmd_timeout)
            else:
                raise SshConnectionError

            self.__p_sftp = self.__p_client.open_sftp()

        except socket.timeout as err:
            raise SshConnectionTimedOut(err)

        except Exception as err:
            raise SshConnectionFailure(err)

    def disconnect(self):
        """ This method disconnect from remote machine """
        self.__p_client.close()
        self.__p_sftp.close()

    def run_command(self, command: str):
        """
        This method run command
        :param command:
        :return: command response or Exception
        """
        try:
            pre_command = '. ~/.bash_profile;'
            stdin, stdout, stderr = self.__p_client.exec_command(f'{pre_command}\n {command}')
            opt = stdout.readlines()
            message = "".join(opt)
            return message
        except Exception as err:
            raise RunCommandError(err)

    def run_background_command(self, command: str):
        """
        This method run command in background '> /dev/null 2>&1 &'
        :param command:
        :return: command response or Exception
        """
        try:
            channel = self.__p_client.invoke_shell()
            out = channel.recv(9999)
            pre_command = '. ~/.bash_profile;'
            channel.send(f'{pre_command}\n {command}')
        except Exception as err:
            raise RunCommandError(err)

    def mkdir(self, remote_path):
        """
        This method mkdir on remote path
        :param remote_path:
        :return:
        """
        try:
            self.__p_sftp.chdir(remote_path)  # Test if remote path exists
        except IOError:
            self.__p_sftp.mkdir(remote_path)  # Create remote path
            self.__p_sftp.chdir(remote_path)

    def put_remote_dir(self, local, remote):
        """
        This method uploads the contents of the local directory to the remote path. The
        local directory needs to exists. All subdirectories in py_perf-code are
        created under target.
        :param local: local path
        :param remote: remote path
        :return:
        """
        if os.path.isdir(local):
            for file_name in os.listdir(local):
                if os.path.isfile(os.path.join(local, file_name)):
                    self.put_remote_file(local, remote, file_name)
                else:
                    self.mkdir(remote, ignore_existing=True)
                    self.put_remote_file(local, remote, file_name)
        else:
            raise PathNotExist(f'Local path does not exist: {local}')

    def get_remote_dir(self, remote, local):
        """
        This method downloads the contents of the remote directory to the local path. The
        remote directory needs to exists. All subdirectories in py_perf-code are
        created under target.
        :param local: local path
        :param remote: remote path
        :return:
        """

        if self.exist(remote_path=remote):
            for file_name in os.listdir(local):
                if os.path.isfile(os.path.join(local, file_name)):
                    self.get_remote_file(remote, local, file_name)
                else:
                    self.mkdir(remote, ignore_existing=True)
                    self.get_remote_file(remote, local, file_name)
        else:
            raise PathNotExist(f'remote path does not exist: {remote}')

    def put_remote_file(self, local, remote, file_name):
        """
        This method upload file name from local to remote path
        :param local:
        :param remote:
        :param file_name:
        :return:
        """
        if os.path.isfile(f'{local}/{file_name}'):
            if self.exist(remote_path=f'{remote}/{file_name}'):
                try:
                    self.__p_sftp.put(f'{local}/{file_name}', f'{remote}/{file_name}')
                except Exception as err:
                    raise SFTPException(err)
            else:
                raise FileNotExist(f'Remote file does not exist: {remote}/{file_name}')
        else:
            raise FileNotExist(f'Local file does not exist: {local}/{file_name}')

    def get_remote_file(self, remote, local, file_name):
        """
        This method download file name from remote to local path
        :param local:
        :param remote:
        :param file_name:
        :return:
        """
        if self.exist(remote_path=f'{remote}/{file_name}'):
            if os.path.isfile(os.path.join(local, file_name)):
                self.__p_sftp.get(f'{remote}/{file_name}', f'{local}/{file_name}')
            else:
                raise FileNotExist(f'Local file does not exist: {local}/{file_name}')
        else:
            raise FileNotExist(f'Remote file does not exist: {remote}/{file_name}')

    def replace_parameter(self, remote_path, file_name, parameter, value, all_line=False):
        """
        This method replaces a parameter in remote file (sed "s@home/test=.@home/test=8@g" test.txt)
        :param remote_path: path to file
        :param file_name: file name
        :param parameter: the parameter
        :param value: the new value
        :param all_line: replace all the line that contains the parameter with value
        :return:
        """
        if all_line:
            res = False
            # This finds a safe delimiter character for sed
            for c in ['/', '@', '#', '$', '%', '^', '&', '*', '-', '<', '>', '!', '~', '+', '-']:
                if c not in parameter:
                    res = True
                    self.run_command(f'sed -i \'s{c}{parameter}{c}{value}{c}g\' {remote_path}/{file_name}')
                    break
            if not res:
                raise RunCommandError
        else:
            if os.path.splitext(file_name)[1] == '.xml':
                self.run_command(f'sed -i \'s/{parameter}[^ ]*"/{parameter}{value}/g\' {remote_path}/{file_name}')
            else:
                self.run_command(f'sed -i \'s/{parameter}.*/{parameter} {value}/g\' {remote_path}/{file_name}')

    def exist(self, remote_path):
        """
        This method check if file or dir exists on remote
        :param: remote_path to file or dir:
        :return: True if exists otherwise False
        """
        try:
            self.__p_sftp.stat(remote_path)
            return True
        except IOError:
            return False

    def rename(self, old_remote_path, new_remote_path):
        """
        This method rename remote dir
        :param old_remote_path:
        :param new_remote_path:
        """
        if self.exist(old_remote_path):
            self.__p_sftp.rename(old_remote_path, new_remote_path)
        else:
            raise PathNotExist(f'remote path does not exist: {old_remote_path}')

    def rmdir(self, remote_path):
        """
        This method delete entire remote file
        """
        if self.exist(remote_path):
            remote_path_files = self.__p_sftp.listdir(path=remote_path)
            for file in remote_path_files:
                self.__p_sftp.remove(f'{remote_path}/{file}')

    def copy(self, remote_source, remote_target):
        """
        This method copy from remote source to target folder
        """
        if self.exist(remote_source):
            if "'" in remote_source or "'" in remote_target:
                raise IllegalFilename("File name with ' ")
            self.run_command(f"cp -arf -- {remote_source}' '{remote_target}'")




import json
from datetime import datetime, timedelta

from benchmark_runner.main.environment_variables import environment_variables
from benchmark_runner.common.elasticsearch.elasticsearch_operations import ElasticSearchOperations


class UpdateGrafanaLatestValueMappings:
    """
    This class updates grafana dashboard with last value mappings from ElasticSearch
    """
    # Prevent error messages from being saved instead of versions in case of a connection issue
    MAX_VERSION_LEN = 20
    LAST_ES_FETCH_DAYS = 30

    def __init__(self, main_libsonnet_path: str):
        # Stamp start/ end value mapping for extract value mapping
        self.start_value_mapping = "//START_VALUE_MAPPING_227"
        self.end_value_mapping = "//END_VALUE_MAPPING_227"
        self._environment_variables_dict = environment_variables.environment_variables_dict
        self.elasticsearch = ElasticSearchOperations(es_host=self._environment_variables_dict.get('elasticsearch', ''),
                                                     es_port=self._environment_variables_dict.get('elasticsearch_port', ''),
                                                     es_user=self._environment_variables_dict.get('elasticsearch_user', ''),
                                                     es_password=self._environment_variables_dict.get('elasticsearch_password', ''))
        self.main_libsonnet_path = main_libsonnet_path
        self.value_mappings = self.get_value_mappings()

    def get_last_elasticsearch_versions(self, last_es_fetch_days=LAST_ES_FETCH_DAYS):
        """
        This method fetches new versions from ElasticSearch
        :param last_es_fetch_days: default fetch 30 days
        :return:
        """
        new_versions = {}
        current_datetime = datetime.now()
        start_datetime = current_datetime - timedelta(days=last_es_fetch_days)

        ids = self.elasticsearch.get_index_ids_between_dates(index='ci-status', start_datetime=start_datetime,
                                                             end_datetime=current_datetime)

        display_versions = ['ocp_version', 'cnv_version', 'kata_version', 'kata_rpm_version', 'odf_version']
        for id in ids:
            data = self.elasticsearch.get_elasticsearch_index_by_id(index='ci-status', id=id)
            for version, value in data['_source'].items():
                if version in display_versions and value not in new_versions.values() and len(value) < self.MAX_VERSION_LEN:
                    # Display only numbers in the Grafana panel, removing characters
                    new_versions[value.translate(str.maketrans('', '', '.-rcef'))] = value

        return new_versions

    def get_value_mappings(self):
        """
        This method gets value mappings from main.libsonnet
        :return:
        """
        # Initialize a flag to indicate whether the lines are within the desired range
        within_range = False
        mapping_lines = []

        # Read the content of the file and store lines between start and end patterns in a list
        with open(self.main_libsonnet_path, 'r') as file:
            for line in file:
                if self.start_value_mapping in line.strip():
                    within_range = True
                    continue  # Skip storing the start line
                elif self.end_value_mapping in line.strip():
                    within_range = False
                    continue  # Skip storing the end line
                if within_range:
                    mapping_lines.append(line.strip())

        # Join the lines to create the final string
        content_string = ''.join(mapping_lines)

        # Convert the string into a dictionary using json.loads()
        return json.loads(content_string)

    def update_value_mappings_last_versions(self, last_versions: dict):
        """
        This method updated value mapping with last versions
        :param last_versions:
        :return:
        """
        # Sort the dictionary by "index" values and convert it to a list of tuples
        sorted_mapping = sorted(self.value_mappings.items(), key=lambda x: x[1]['index'])
        # Get the maximum index value and the corresponding key
        max_index, max_key = sorted_mapping[-1][1]['index'], sorted_mapping[-1][0]

        num = 1
        for key, value in last_versions.items():
            if not self.value_mappings.get(key):
                self.value_mappings[key] = {"index": int(max_index) + num, "text": value}
                num += 1

    def update_main_libsonnet(self):
        """
        This method updated last versions in main.libsonnet
        :return:
        """
        # Read the content of the file and store it in a list
        with open(self.main_libsonnet_path, 'r') as file:
            lines = file.readlines()
        # Find the indices of the lines between start and end
        start_index = None
        end_index = None
        for i, line in enumerate(lines):
            if self.start_value_mapping in line:
                start_index = i
            elif self.end_value_mapping in line:
                end_index = i

        # Replace the lines between start and end with the dictionary
        if start_index is not None and end_index is not None:
            lines[start_index + 1:end_index] = [f'\t\t\t\t {json.dumps(self.value_mappings, indent=None)} \n']

        # Write the modified content back to the file
        with open(self.main_libsonnet_path, 'w') as file:
            file.writelines(lines)




class ClusterBusterError(Exception):
    """ Base class for all benchmark operator error classes.
        All exceptions raised by the benchmark runner library should inherit from this class. """
    pass


class MissingResultReport(ClusterBusterError):
    """
    This class is error for missing cluster buster report result
    """
    def __init__(self):
        self.message = "Missing cluster buster result report"
        super(MissingResultReport, self).__init__(self.message)


class MissingElasticSearch(ClusterBusterError):
    """
    This class is error for missing ElasticSearch details
    """
    def __init__(self):
        self.message = "Missing ElasticSearch details"
        super(MissingElasticSearch, self).__init__(self.message)




import json
# import urllib library
from urllib.request import urlopen
from urllib.error import URLError, HTTPError

from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger


class AssistedInstallerVersions:
    """
    This class get the latest assisted installer version from https://openshift-release.apps.ci.l2s4.p1.openshiftapps.com/graph
    """

    OCP_VERSIONS_URL = "https://openshift-release.apps.ci.l2s4.p1.openshiftapps.com/graph"
    # MAJOR.MINOR.PATCH.BUILD
    IND_MAJOR = 0
    IND_MINOR = 1
    IND_PATCH = 2
    IND_BUILD = 3

    def __init__(self):
        self.__url = self.OCP_VERSIONS_URL

    def __get_versions(self):
        """
        This method get the OCP versions from url and return json data
        :return: versions in json
        """
        response = {}
        # store the response of URL
        try:
            response = urlopen(self.__url)
        except HTTPError as err:
            logger.error('HTTPError code: ', err.code)
        except URLError as err:
            logger.error('URLError code: ', err.reason)
        # Storing the JSON response from url in data
        if response:
            return json.loads(response.read())
        else:
            raise Exception(f'Error reading OCP versions from url: {self.__url}')

    def get_several_latest_versions(self, latest_versions: list):
        """
        The method returns list of latest versions
        @param latest_versions: list of required latest versions
        @return:
        """
        result = []
        for latest_version in latest_versions:
            latest = self.get_latest_version(latest_version)
            result.append(latest)
        return result

    def get_latest_version(self, latest_version: str):
        """
        This method returns the latest version from json data
        :param latest_version: 4.XX or 4.XX.0-rc/ec/fc
        rc=release candidate| ec=engineering candidate| fc=feature candidate
        :return:
        """
        release_list = []
        release_candidate_list = []
        release_version = ''  # Version=4.X.X
        release_candidate_version = ''  # Version=4.X.0-rc.X
        latest_version = latest_version.split('.')
        for version in self.__get_versions()['nodes']:
            version_data = str(version['version']).split('.')
            if f"{latest_version[self.IND_MAJOR]}.{latest_version[self.IND_MINOR]}" == f'{version_data[self.IND_MAJOR]}.{version_data[self.IND_MINOR]}':
                # Release version=4.X.X
                if len(version_data) == self.IND_PATCH + 1:
                    release_version = f'{version_data[self.IND_MAJOR]}.{version_data[self.IND_MINOR]}'
                    release_list.append(int(version_data[self.IND_PATCH]))
                # Release candidate version=4.X.0-rc.X and skip 4.X.X-0.nightly-2022-09-06-225330
                if len(version_data) == self.IND_BUILD + 1 and len(version_data[self.IND_BUILD]) <= 2:
                    # Specific Release candidate version=4.X.0-rc
                    if len(latest_version) == self.IND_BUILD:
                        if latest_version[self.IND_PATCH] == version_data[self.IND_PATCH]:
                            release_candidate_version = '.'.join([version_data[self.IND_MAJOR], version_data[self.IND_MINOR], version_data[self.IND_PATCH]])
                            release_candidate_list.append(int(version_data[self.IND_BUILD]))
                    else:
                        release_candidate_version = '.'.join(
                            [version_data[self.IND_MAJOR], version_data[self.IND_MINOR], version_data[self.IND_PATCH]])
                        release_candidate_list.append(int(version_data[self.IND_BUILD]))

        # Priority to release version
        if release_list and len(latest_version) == self.IND_PATCH:
            return f'{release_version}.{max(release_list)}'.strip()
        elif release_candidate_list:
            return f'{release_candidate_version}.{max(release_candidate_list)}'.strip()



import os
import time
from datetime import datetime
from typeguard import typechecked
import ast  # change string list to list

from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger, datetime_format
from benchmark_runner.main.environment_variables import environment_variables
from benchmark_runner.common.remote_ssh.remote_ssh import ConnectionData, RemoteSsh
from benchmark_runner.common.ocp_resources.create_ocp_resource import CreateOCPResource
from benchmark_runner.common.oc.oc import OC
from benchmark_runner.common.github.github_operations import GitHubOperations
from benchmark_runner.common.clouds.BareMetal.bare_metal_exceptions import MissingMasterNodes, MissingWorkerNodes, OCPInstallationFailed, OCPUpgradeFailed
from benchmark_runner.common.ssh.ssh import SSH
from benchmark_runner.common.assisted_installer.assisted_installer_latest_version import AssistedInstallerVersions


class BareMetalOperations:
    """
    This class is responsible for Bare metal OpenShift installation operations, all commands run on remote provision server
    """
    LATEST = 'latest'
    SHORT_TIMEOUT = 600

    @typechecked
    def __init__(self, user: str):
        self._environment_variables_dict = environment_variables.environment_variables_dict
        self._user = user
        self._ocp_env_flavor = self._environment_variables_dict.get('ocp_env_flavor', '')
        self._create_pod_ci_cmd = self._environment_variables_dict.get('create_pod_ci_cmd', '')
        self._provision_kubeadmin_password_path = self._environment_variables_dict.get('provision_kubeadmin_password_path', '')
        self._provision_kubeconfig_path = self._environment_variables_dict.get('provision_kubeconfig_path', '')
        self._provision_installer_path = self._environment_variables_dict.get('provision_installer_path', '')
        self._provision_installer_cmd = self._environment_variables_dict.get('provision_installer_cmd', '')
        self._provision_installer_log = self._environment_variables_dict.get('provision_installer_log', '')
        self._install_ocp_version = self._environment_variables_dict.get('install_ocp_version', '')
        self._cluster_type = self._environment_variables_dict.get('cluster_type', '')
        self._expected_nodes = self._environment_variables_dict.get('expected_nodes', '')
        if self._expected_nodes:
            # string to dict
            self._expected_nodes = ast.literal_eval(self._expected_nodes)
        self._ocp_version_build = self._environment_variables_dict.get('ocp_version_build', '')
        self._num_odf_disks = int(self._environment_variables_dict.get('num_odf_disk', 1))
        self._provision_ip = self._environment_variables_dict.get('provision_ip', '')
        self._provision_port = int(self._environment_variables_dict.get('provision_port', ''))
        self._provision_timeout = int(self._environment_variables_dict.get('provision_timeout', ''))
        self._container_private_key_path = self._environment_variables_dict.get('container_private_key_path', '')
        self._connection_data = ConnectionData(host_name=self._provision_ip,
                                                user_name=self._user,
                                                port=self._provision_port,
                                                timeout=self._provision_timeout,
                                                ssh_key=self._container_private_key_path)
        self._remote_ssh = RemoteSsh(self._connection_data)
        if self._environment_variables_dict.get('github_repository_short', ''):
            self._github_operations = GitHubOperations()
        self._ssh = SSH()
        self._cli = self._environment_variables_dict.get('cli', '')
        self._upgrade_ocp_version = self._environment_variables_dict.get('upgrade_ocp_version', '')

    def _get_kubeadmin_password(self):
        """
        This method returns kubeadmin password if exist
        :return:
        """
        if self._remote_ssh.exist(remote_path=self._provision_kubeadmin_password_path):
            return self._remote_ssh.run_command(f'cat {self._provision_kubeadmin_password_path}')

    def _get_kubeconfig(self):
        """
        This method returns kubeconfig if exist
        :return:
        """
        if self._remote_ssh.exist(remote_path=self._provision_kubeconfig_path):
            return self._remote_ssh.run_command(f'cat {self._provision_kubeconfig_path}')

    def _wait_for_install_complete(self, sleep_time: int = SHORT_TIMEOUT):
        """
        This method waits till ocp install complete
        :param sleep_time:
        :return:
        """
        current_wait_time = 0
        install_log = ''
        while self._provision_timeout <= 0 or current_wait_time <= self._provision_timeout:
            install_log = self._remote_ssh.run_command(self._provision_installer_log)
            if 'failed=0' in install_log:
                return True
            elif 'failed=1' in install_log:
                return False
            logger.info(f'Waiting till OCP install complete, waiting {int(current_wait_time/60)} minutes')
            # sleep for x seconds
            time.sleep(sleep_time)
            current_wait_time += sleep_time
        raise OCPInstallationFailed(install_log)

    def _wait_for_upgrade_complete(self, oc: OC, sleep_time: int = SHORT_TIMEOUT):
        """
        This method waits till ocp upgrade complete
        :param sleep_time:
        :return:
        """
        current_wait_time = 0
        logger.info(f'Waiting until the upgrade to version {self._upgrade_ocp_version} starts...')
        oc.wait_for_upgrade_start(upgrade_version=self._upgrade_ocp_version)
        while self._provision_timeout <= 0 or current_wait_time <= self._provision_timeout and oc.upgrade_in_progress():
            logger.info(f'Waiting till OCP upgrade complete, waiting {int(current_wait_time / 60)} minutes')
            # sleep for x seconds
            time.sleep(sleep_time)
            current_wait_time += sleep_time
        if f'Cluster version is {self._upgrade_ocp_version}' == oc.get_cluster_status():
            return True
        else:
            raise OCPUpgradeFailed(status=oc.get_cluster_status())

    def _install_ocp_cmd(self):
        """
        This method returns local ocp installer cmd if exist
        :return:
        """
        if self._remote_ssh.exist(remote_path=self._provision_installer_path):
            return self._provision_installer_cmd

    @logger_time_stamp
    def connect_to_provisioner(self):
        """
        This method connect to remote provision machine
        :return:
        """
        self._remote_ssh.connect()

    @logger_time_stamp
    def disconnect_from_provisioner(self):
        """
        This method
        :return:
        """
        self._remote_ssh.disconnect()

    def _get_installation_version(self):
        """
        This method return installation version
        :return:
        """
        assisted_installer_versions = AssistedInstallerVersions()
        if self.LATEST in self._install_ocp_version:
            openshift_version_data = self._install_ocp_version.split('-')
            # release candidate or engineering candidate or feature candidate version
            if '-rc' in self._install_ocp_version or '-ec' in self._install_ocp_version or '-fc' in self._install_ocp_version:
                return assisted_installer_versions.get_latest_version(latest_version=f'{openshift_version_data[1]}-{openshift_version_data[2]}')
            # release version
            else:
                return assisted_installer_versions.get_latest_version(latest_version=openshift_version_data[1])
        else:
            return self._install_ocp_version

    def get_ocp_server_version(self):
        """
        This method return ocp server version
        :return:
        """
        return self._remote_ssh.run_command(command=f"{self._cli} version -ojson | jq -r '.openshiftVersion'").strip()

    def version_already_installed(self):
        """
        This method validate if version is already install
        :return: True if it already installed, False if it NOT already installed
        """
        if self.get_ocp_server_version() == 'null':
            return False
        elif self.LATEST in self._install_ocp_version and self.get_ocp_server_version() == self._get_installation_version():
            return self.get_ocp_server_version().strip()
        elif self._install_ocp_version == self.get_ocp_server_version():
            return self._install_ocp_version
        else:
            return False

    @logger_time_stamp
    def restart_pod_ci(self):
        """
        This method restart pod ci: elastic, kibana, grafana, nginx and flask - solved connectivity issue after installation
        :return:
        """
        self._remote_ssh.run_command('podman pod restart pod_ci')

    @typechecked()
    @logger_time_stamp
    def update_ocp_version(self, file_name: str):
        """
        This method update the ocp version on provision machine
        :param file_name:
        :return:
        """
        # Get the latest assisted installer version
        if self.LATEST in self._install_ocp_version:
            self._install_ocp_version = self._get_installation_version()
        openshift_version_data = self._install_ocp_version.split('.')
        self._remote_ssh.replace_parameter(remote_path='/root/jetlag/ansible/vars',
                                            file_name=file_name,
                                            parameter='ocp_release_image:',
                                            value=fr'quay.io\/openshift-release-dev\/ocp-release:{self._install_ocp_version}-x86_64')
        self._remote_ssh.replace_parameter(remote_path='/root/jetlag/ansible/vars',
                                            file_name=file_name,
                                            parameter='openshift_version:',
                                            value=f'"{openshift_version_data[0]}.{openshift_version_data[1]}"')

    def update_provision_config(self):
        """
        This method update provision config file
        :return:
        """
        # update ssh config - must add it and also mount it
        self._ssh.run(f"echo Host provision > /{self._user}/.ssh/config")
        self._ssh.run(f"echo -e '\t'HostName {self._provision_ip} >> /{self._user}/.ssh/config")
        self._ssh.run(f"echo -e '\t'User {self._user} >> /{self._user}/.ssh/config")
        self._ssh.run(f"echo -e '\t'IdentityFile {self._container_private_key_path} >> /{self._user}/.ssh/config")
        self._ssh.run(f"echo -e '\t'StrictHostKeyChecking no >> /{self._user}/.ssh/config")
        self._ssh.run(f"echo -e '\t'ServerAliveInterval 30 >> /{self._user}/.ssh/config")
        self._ssh.run(f"echo -e '\t'ServerAliveCountMax 5 >> /{self._user}/.ssh/config")
        self._ssh.run(f"chmod 600 /{self._user}/.ssh/config")

    @logger_time_stamp
    def run_ocp_installer(self):
        """
        This method runs ocp assisted installer
        :return: True if installation success and raise exception if installation failed
        """
        self.update_provision_config()
        logger.info(f'Starting OCP assisted installer, Start time: {datetime.now().strftime(datetime_format)}')
        # Must add -t otherwise remote ssh of ansible will not end
        self._ssh.run(f"ssh -t provision \"{self._install_ocp_cmd()}\" ")
        logger.info(f'OpenShift cluster {self._get_installation_version()} version is installed successfully, End time: {datetime.now().strftime(datetime_format)}')

    @logger_time_stamp
    def run_ocp_upgrade(self, oc: OC):
        """
        This method runs ocp upgrade
        @param oc:
        :return: True if installation success and raise exception if installation failed
        """
        logger.info(f'Starting OCP upgrade, Start time: {datetime.now().strftime(datetime_format)}')
        logger.info(f'Stop OCP healthcheck')
        oc.healthcheck(action='stop')
        oc.upgrade_ocp(upgrade_ocp_version=self._upgrade_ocp_version)
        self.verify_upgrade_complete(oc=oc)
        logger.info(f'Resume OCP healthcheck')
        oc.healthcheck(action='resume')
        logger.info(f'Ending OCP upgrade, End time: {datetime.now().strftime(datetime_format)}')

    @logger_time_stamp
    def verify_install_complete(self):
        """
        This method verifies that the installation has completed
        :return: Returns True if the installation is successful; raises an exception if the installation fails
        """
        complete = self._wait_for_install_complete()
        if not complete:
            install_log = self._remote_ssh.run_command(self._provision_installer_log)
            raise OCPInstallationFailed(install_log)
        else:
            return True

    def is_cluster_upgraded(self, oc: OC, cnv_version: str, odf_version: str, lso_version: str):
        """
        This method checks if cluster/operators upgraded successfully
        @param oc:
        @param cnv_version:
        @param odf_version:
        @param lso_version:
        @return:
        """
        # check cluster version upgrade
        if oc.get_upgrade_version() == self._upgrade_ocp_version:
            lso_upgraded = oc.wait_for_operator_installation(operator='lso', version=lso_version, namespace='openshift-local-storage')
            odf_upgraded = oc.wait_for_operator_installation(operator='odf', version=odf_version, namespace='openshift-storage')
            cnv_upgraded = oc.wait_for_operator_installation(operator='cnv', version=cnv_version, namespace='openshift-cnv')
            # check operators versions upgrade
            if cnv_upgraded and lso_upgraded and odf_upgraded:
                return True
        return False

    @logger_time_stamp
    def verify_upgrade_complete(self, oc: OC):
        """
        This method verifies that the upgrade has completed
        :return: True if the upgrade is successful; raises an exception if the upgrade fails
        """
        complete = self._wait_for_upgrade_complete(oc)
        if complete:
            return True
        raise OCPUpgradeFailed(status=oc.get_cluster_status())

    @logger_time_stamp
    def oc_login(self):
        """
        This method login to the cluster with new credentials
        :return:
        """
        oc = OC(kubeadmin_password=self._get_kubeadmin_password())
        oc.login()
        return oc

    @staticmethod
    def get_missing_node(expected_nodes: list, actual_nodes: list):
        """
        This method returns missing nodes
        :return:
        """
        return set(expected_nodes) - set(actual_nodes)

    @logger_time_stamp
    def verify_cluster_is_up(self, oc: OC):
        """
        This method verify that master/worker nodes are up and running
        :return:
        """
        master_nodes = oc.get_master_nodes()
        if len(master_nodes.split()) == len(self._expected_nodes['master']):
            logger.info('master nodes are up and running')
            if self._ocp_env_flavor == 'PERF' and not self._cluster_type == 'SNO':
                worker_nodes = oc.get_worker_nodes()
                # if there are worker nodes, they must be 3
                if len(worker_nodes.split()) > 1:
                    if len(worker_nodes.split()) == len(self._expected_nodes['worker']):
                        logger.info('worker nodes are up and running')
                    else:
                        missing_node = self.get_missing_node(expected_nodes=self._expected_nodes['worker'], actual_nodes=worker_nodes.split())
                        raise MissingWorkerNodes(missing_node)
        else:
            missing_node = self.get_missing_node(expected_nodes=self._expected_nodes['master'], actual_nodes=master_nodes.split())
            raise MissingMasterNodes(missing_node)

    @staticmethod
    @logger_time_stamp
    @typechecked
    def install_ocp_resources(resources: list, upgrade_version: str = None):
        """
        This method installs OCP resources 'cnv', 'lso', 'odf'
        :param resources:
        :param upgrade_version: upgrade version, default None
        :return:
        """
        create_ocp_resource = CreateOCPResource()
        for resource in resources:
            create_ocp_resource.create_resource(resource=resource, upgrade_version=upgrade_version)

    @logger_time_stamp
    def update_ocp_github_credentials(self):
        """
        This method updates GitHub secrets kubeconfig and kubeadmin_password
        :return:
        """
        self._github_operations.create_secret(secret_name=f'{self._ocp_env_flavor}_KUBECONFIG', unencrypted_value=self._get_kubeconfig())
        self._github_operations.create_secret(secret_name=f'{self._ocp_env_flavor}_KUBEADMIN_PASSWORD',
                                              unencrypted_value=self._get_kubeadmin_password())

    def get_ocp_install_time(self):
        """
        This method returns installation time
        :param self:
        :return:
        """
        install_log = self._remote_ssh.run_command(self._provision_installer_log)
        return install_log.split()[-1].strip('"')
